#include "global.h"


/*******************************************************************************
                     Mapping  Function 
 *******************************************************************************/


void map_to_pnum(grid, st, pn)
     int grid[Rankmax];
     int st;
     int *pn;
{
  int *nproc;
  int i, mult=1;
  int rank;

  *pn = 0;

  if ( st == SOURCE ) {
    rank = s_rank;
    nproc = s_nproc;
  }
  else {
    rank = t_rank;
    nproc = t_nproc;
  }
    
  for (i=0; i<rank; ++i) {
    *pn += grid[i] * mult;
    mult *= nproc[i];
  };

};



void map_to_grid(st, pn, grid)
     int pn;
     int st;
     int grid[Rankmax];
{
  int *nproc;
  int i, mult, mod;
  int rank;

  if ( st == SOURCE ) {
    rank = s_rank;
    nproc = s_nproc;
    mult  = s_tot_nproc;
  }
  else {
    rank = t_rank;
    nproc = t_nproc;
    mult  = t_tot_nproc;
  };
  
  mod = pn;

  for (i=rank-1; i>=0; --i) {
    mult /= nproc[i];
    grid[i] = mod / mult;
    mod %= mult;
  }

};


Lindex map_to_local(st, i, d)
     int st;
     Gindex i;
     int d;
{
  int n, P, b;

  if ( st == SOURCE ) {
    n = s_N[d] - 1;
    b = s_width[d];
    P = s_nproc[d];
  }
  else {
    n = t_N[d] - 1;
    b = t_width[d];
    P = t_nproc[d];
  };
 
  return ( (i / (P*b))*b + i % b );
}


Gindex map_to_global(st, p, l, d)
     int st;
     int p;
     Lindex l;
     int d;
{
  int P, b;

  if ( st == SOURCE ) {
    b = s_width[d];
    P = s_nproc[d];
  }
  else {
    b = t_width[d];
    P = t_nproc[d];
  };
 
  return ( (l / b) * b*P + b*p + l%b );
}


int map_to_proc(st, i, d)
     int    st;
     Gindex i;
     int    d;
{
  int P, b;

  if ( st == SOURCE ) {
    b = s_width[d];
    P = s_nproc[d];
  }
  else {
    b = t_width[d];
    P = t_nproc[d];
  };

  return ( ( i / b ) % P );
}


/*******************************************************************************
                     Utility Function 
 *******************************************************************************/



/*============================================*
 *  Integer Function : Clog2(n)               *
 *                                            *
 *   - Input  = positive integer              *
 *   - Return = Ceiling ( log2(n) )           *
 *============================================*/

int Clog2(n)
     int n;
{
  int i = 0;
  int n1 = n;

  if ( n < 1 ) return -1;

  while ( n1 >>= 1 )
    ++i;

  if ( 1<<i != n )
    return i+1;
  else
    return i;
}


/*============================================*
 *  Boolean Function : Blog2(n)               *
 *                                            *
 *   - Input  = positive integer              *
 *   - Return = true if log2(n) is integer    *
 *============================================*/

Boolean Blog2(n)
     int n;
{
  int x = n, m = 0;
  while ( n > 1 ) {
     m += 1;
     n /= 2;
   };

  return ( x == ( 1 << m ) );
};


/*============================================*
 *  Integer Function : Ilog2(n)               *
 *                                            *
 *   - Input  = positive integer              *
 *   - Return = log2(n) if log2(n) is integer *
 *              otherwise return -1           *
 *============================================*/

int Ilog2(n)
     int n;
{
  int i = 0;
  int n1 = n;

  if ( n < 1 ) return -1;

  while ( n1 >>= 1 )
    ++i;

  if ( 1<<i != n )
    return -1;
  else
    return i;
}



/*============================================*
 *  Integer Function : pow2(x)                *
 *                                            *
 *   - Input  = exponential x                 *
 *   - Return = 2^x                           *
 *============================================*/

int pow2(x)
     int x;
{
  int p=1;
  int i;

  for (i=0; i<x; ++i)
    p <<= 1;

  return p;
}

