#include "global.h"
#include  "raw_type.h"
 
extern int pr_control;

MPI_Status status;

int tag = 0;


/******************************************************************************
 *                                                                            *
 *  Function : print_array1()                                                 *
 *                                                                            *
 ******************************************************************************/

void print_array1(stp, array, scheme, type)
     int stp;
     Data_Type array[Lmax];
     char *scheme;
     char type;
{
  int r, /*global*/
      i; /*local*/
  int pn;
  Data_Type ele;
  int st;
  int size;

  MPI_Barrier(WORLD);
  if (only_master) {
    if (stp == SOURCE) 
      printf("\n *************** %s: SOURCE ARRAY [Type=%c] *****************\n",
           scheme, type); 
    else if (stp == POINTER)
      printf("\n ******************** %s: POINTER ARRAY *******************\n",
  	     scheme); 
    else
      printf("\n *************** %s: TARGET ARRAY [Type=%c] *****************\n",
           scheme, type); 
  }

  if (stp == TARGET) {
    st = TARGET;
    size = t_N[0];
  }
  else {
    st = SOURCE;
    size = s_N[0];
  }

  MPI_Barrier(WORLD); 

  for (r=0; r<size; ++r) {
    MPI_Barrier(WORLD); 
    
    i = map_to_local(st, r, 0);
    pn = map_to_proc(st, r, 0);

    ele = array[i];

    if ( my_pnum == 0 ) {

      if ( pn == 0 ) {
	if (type = 'i' )
	  printf("%7d", ele);
	else
	  printf("%7.0f", ele);
      }
      else {
	MPI_Recv( &ele, sizeof(Data_Type), MPI_BYTE, pn, tag, WORLD, &status );
	if (type = 'i' )
	  printf("%7d", ele);
	else
	  printf("%7.0f", ele);
      };
    }
    else if ( my_pnum == pn ) 
      MPI_Send( &ele, sizeof(Data_Type), MPI_BYTE, master, tag, WORLD );
  }

  if (only_master)
    printf("\n\n");
       
  MPI_Barrier(WORLD);

}


/******************************************************************************
 *                                                                            *
 *  Function : print_array1_complex()
 *                                                                            *
 ******************************************************************************/

void print_array1_complex(stp, array, scheme, type)
     int stp;
     COMPLEX_Data_Type array[Lmax];
     char *scheme;
     char type;
{
  int r, /*global*/
      i; /*local*/
  int pn;
  float ele;
  int st;
  int size;

  MPI_Barrier(WORLD);
  if (only_master) {
    if (stp == SOURCE) 
      printf("\n *************** %s: SOURCE ARRAY [Type=%c] *****************\n",
           scheme, type); 
    else if (stp == POINTER)
      printf("\n ******************** %s: POINTER ARRAY *******************\n",
  	     scheme); 
    else
      printf("\n *************** %s: TARGET ARRAY [Type=%c] *****************\n",
           scheme, type); 
  }

  if (stp == TARGET) {
    st = TARGET;
    size = t_N[0];
  }
  else {
    st = SOURCE;
    size = s_N[0];
  }

  MPI_Barrier(WORLD); 

  /*----------------------- real part ---------------------------------*/
  for (r=0; r<size; ++r) {
    MPI_Barrier(WORLD); 
    
    i = map_to_local(st, r, 0);
    pn = map_to_proc(st, r, 0);

    ele = array[i].Real;

    if ( my_pnum == 0 ) {

      if ( pn == 0 ) {
	printf("%7.0f", ele);
      }
      else {
	MPI_Recv( &ele, sizeof(float), MPI_BYTE, pn, tag, WORLD, &status );
	printf("%7.0f", ele);
      };
    }
    else if ( my_pnum == pn ) 
      MPI_Send( &ele, sizeof(float), MPI_BYTE, master, tag, WORLD );
  }

  if (only_master)
    printf("\n\n\n\n");
       
  MPI_Barrier(WORLD);

  /*----------------------- imag part ---------------------------------*/
  for (r=0; r<size; ++r) {
    MPI_Barrier(WORLD); 
    
    i = map_to_local(st, r, 0);
    pn = map_to_proc(st, r, 0);

    ele = array[i].Imag;

    if ( my_pnum == 0 ) {

      if ( pn == 0 ) {
	printf("%7.0f", ele);
      }
      else {
	MPI_Recv( &ele, sizeof(float), MPI_BYTE, pn, tag, WORLD, &status );
	printf("%7.0f", ele);
      };
    }
    else if ( my_pnum == pn ) 
      MPI_Send( &ele, sizeof(float), MPI_BYTE, master, tag, WORLD );
  }

  if (only_master)
    printf("\n\n");
       
  MPI_Barrier(WORLD);

}


/******************************************************************************
 *                                                                            *
 *  Function : print_array2(array, kind)
 *                                                                            *
 ******************************************************************************/


void print_array2(stp, array, scheme)
     int stp;
     Data_Type array[Lmax][Lmax];
     char *scheme;
{
  int r, c,     /* global index */
      i, j,     /* local index  */
      pn, grid[Rankmax];
  int ele, st;
  int size0, size1;

  MPI_Barrier(WORLD);

  if (only_master) {
    if (stp == SOURCE) 
      printf("\n ******************** %s: SOURCE ARRAY ********************\n",
  	     scheme); 
    else if (stp == POINTER)
      printf("\n ******************** %s: POINTER ARRAY *******************\n",
  	     scheme); 
    else
      printf("\n ******************** %s: TARGET ARRAY ********************\n",
  	     scheme); 
  }

  if (stp == TARGET) {
    st = TARGET;
    size0 = t_N[0];
    size1 = t_N[1];
  }
  else {
    st = SOURCE;
    size0 = s_N[0];
    size1 = s_N[1];
  }

   fflush(stdout);

   MPI_Barrier(WORLD); 

   for (r=0; r<size1; ++r) {
      MPI_Barrier(WORLD); 
      
      i = map_to_local(st, r, 1);
      grid[1] = map_to_proc(st, r, 1);

      for (c=0; c<size0; ++c) {  
         MPI_Barrier(WORLD);

         j  = map_to_local(st, c, 0);
         grid[0] = map_to_proc(st, c, 0);

         map_to_pnum(grid, st, &pn);

         ele = array[i][j];

         MPI_Barrier(WORLD);
         if ( my_pnum == 0 ) {

            if ( pn == 0 )
              printf("%7d", ele);
            else {
	      MPI_Recv( &ele, sizeof(Data_Type), MPI_BYTE, pn, tag, WORLD, &status );
              printf("%7d", ele);
           };
         }
         else if ( my_pnum == pn ) 
	   MPI_Send( &ele, sizeof(Data_Type), MPI_BYTE, master, tag, WORLD );
       
      }; /* for(c) */

         if (my_pnum == 0 ) {
              printf("\n");
           };
      
   }; /* for(r) */

   MPI_Barrier(WORLD);
   if (only_master)
     printf("\n");

 };

 
/******************************************************************************
 *                                                                            *
 *  Function : print_bucket()
 *                                                                            *
 ******************************************************************************/


void print_bucket(array, size, title, st)
     int *array;
     int size;
     char *title;
     int st;
{
  int tot, pn, i;
  int temp[128];

  if (only_master) {
    printf("\n\n ===========>>> %s (size= %d)\n", title, size);
    fflush(stdout);
  }

  MPI_Barrier(WORLD);

  if (st == SOURCE)
    tot = s_tot_nproc;
  else
    tot = t_tot_nproc;

  if (only_master) {
    printf("\n  -----------[PN:%3d]----------- \n", master); 
    for (i=0; i<size; ++i)
      printf("%6d", array[i]);
    fflush(stdout);
  }

  for (pn=1; pn<tot; ++pn) {
    if ( only_master ) {
      MPI_Recv( temp, size, MPI_INT, pn, TAG, WORLD, &status );
      printf("\n  -----------[PN:%3d]----------- \n", pn); 
      for (i=0; i<size; ++i)
	printf("%6d", temp[i]);
      fflush(stdout);
    }

    else if (my_pnum == pn) 
      MPI_Send( array, size, MPI_INT, master, TAG, WORLD );

    MPI_Barrier(WORLD);
  }

  if (only_master)
    printf("\n\n");
       
  MPI_Barrier(WORLD);

}


/******************************************************************************
 *                                                                            *
 *  Function : print_buffer()
 *                                                                            *
 ******************************************************************************/



void print_buffer(array, kind)
     Buffer_Type *array;
     char kind;
{
  int pn, i, size;

  MPI_Barrier(WORLD);

  if (only_master) {
    if (kind == 'S') {
      printf("\n\n ******************** SEND BUFFER ********************\n"); 
      size = tot_sendl/2;
    }
    else  {
      printf("\n\n ******************** RECV BUFFER ********************\n");
      size = tot_recvl/2;
    }
    fflush(stdout);
  }
  else {
    if (kind == 'S') 
      size = tot_sendl/2;
    
    else  
      size = tot_recvl/2;
  }


  MPI_Barrier(WORLD);

  for (pn=0; pn<tot_nproc; ++pn) {

    MPI_Barrier(WORLD);
    fflush(stdout);
    MPI_Barrier(WORLD);
    if ( my_pnum == pn ) {
      printf("\n  ==============> [PN:%3d] Size = %d \n", pn, size*2); 
      for (i=0; i<size; ++i)
	printf("[%3d,%3d]  ", array[i].data, array[i].ptr);
      fflush(stdout);
    }
    MPI_Barrier(WORLD);
    fflush(stdout);
    MPI_Barrier(WORLD);
  }

  MPI_Barrier(WORLD);

  if (only_master)
    printf("\n");

}

/******************************************************************************
 *                                                                            *
 *  Function : print_buffer_raw1()
 *                                                                            *
 ******************************************************************************/



void print_buffer_raw1(array, kind, size)
     int *array;
     char kind;
     int size;
{
  int pn, i;

  MPI_Barrier(WORLD);
  if (only_master) {
    if (kind == 'S') {
      printf("\n\n ******************** SEND BUFFER ********************\n"); 
    }
    else  if (kind == 'R') {
      printf("\n\n ******************** RECV BUFFER ********************\n");
    }
    fflush(stdout);
  }

  MPI_Barrier(WORLD);

  for (pn=0; pn<tot_nproc; ++pn) {

    MPI_Barrier(WORLD);

    if ( my_pnum == pn ) {
      printf("\n  ==============> [PN:%3d] Size = %d \n", pn, size); 
      for (i=0; i<size; ++i)
	printf("[%3d]  ", array[i]);
      fflush(stdout);
    }

    MPI_Barrier(WORLD);
    fflush(stdout);
    MPI_Barrier(WORLD);
  }

  MPI_Barrier(WORLD);
  if (only_master)
    printf("\n");


}


/******************************************************************************
 *                                                                            *
 *  Function : print_buffer_raw2()
 *                                                                            *
 ******************************************************************************/



void print_buffer_raw2(array, kind, size)
     int *array;
     char kind;
     int size;
{
  int pn, i;

  MPI_Barrier(WORLD);
  if (only_master) {
    if (kind == 'S') {
      printf("\n\n ******************** SEND BUFFER ********************\n"); 
    }
    else  if (kind == 'R') {
      printf("\n\n ******************** RECV BUFFER ********************\n");
    }
    fflush(stdout);
  }

  MPI_Barrier(WORLD);

  for (pn=0; pn<tot_nproc; ++pn) {

    MPI_Barrier(WORLD);
    fflush(stdout);
    MPI_Barrier(WORLD);

    if ( my_pnum == pn ) {
      printf("\n  ==============> [PN:%3d] Size = %d \n", pn, size); 
      for (i=0; i<size; ++i)
	printf("[%3d,%3d]  ", array[2*i], array[2*i+1]);
      fflush(stdout);
    }
    MPI_Barrier(WORLD);
    fflush(stdout);
    MPI_Barrier(WORLD);
  }

  MPI_Barrier(WORLD);
  if (only_master) 
    printf("\n");


}



/******************************************************************************
 *                                                                            *
 *  Function : print_comm_vector()
 *                                                                            *
 ******************************************************************************/


void print_comm_vector(array, kind)
     int *array;
     char kind;
{
  int pn, i, size;
  int temp[8];

  MPI_Barrier(WORLD);
  if (only_master) {
    if (kind == 'S') 
      printf("\n\n ***************** SEND VECTOR: self=%d *******************\n",
  	     sendl_self); 
    else  if (kind == 'R')
      printf("\n\n ******************** RECV VECTOR ********************\n");
    else  if (kind == 's')
      printf("\n\n ******************** SEND OFFSET ********************\n");
    else  
      printf("\n\n ******************** RECV OFFSET ********************\n");
    fflush(stdout);
  }

  MPI_Barrier(WORLD);

  if (only_master) {
    printf("\n  ==============> [PN:%3d]", master); 
    if (kind == 'S') 
    	printf(" Sendl-Self = %d\n", sendl_self);
    else
    	printf("\n");
    
    for (i=0; i<tot_nproc; ++i)
    	printf("%4d ", array[i]);
    fflush(stdout);
  }

  for (pn=1; pn<tot_nproc; ++pn) {

    if ( only_master ) {
      MPI_Recv( temp, tot_nproc, MPI_INT, pn, 
		MPI_ANY_TAG, WORLD, &status );
      printf("\n  ==============> [PN:%3d]", pn); 
      
      if (kind == 'S') 
	printf(" Sendl-Self = %d\n", status.MPI_TAG);
      else
	printf("\n");

      for (i=0; i<tot_nproc; ++i)
	printf("%4d ", temp[i]);
      fflush(stdout);
    }

    else if (my_pnum == pn) 
      MPI_Send( array, tot_nproc, MPI_INT, master, sendl_self, WORLD );

    MPI_Barrier(WORLD);

  }

  MPI_Barrier(WORLD);
  if (only_master) 
    printf("\n\n");


}


/******************************************************************************
 *                                                                            *
 *  Function : print_comm_matrix()
 *                                                                            *
 ******************************************************************************/


void print_comm_matrix(array, kind, total, title)
     int *array;
     char kind;
     int total;
     char *title;
{
  int pn, i, size;

  MPI_Barrier(WORLD);
  if (only_master) {
    if (kind == 'D' || kind == 'd') 
      printf("\n\n ***************** Dense Vector: %s  *******************\n", title);
    else  if (kind == 'S' || kind == 's') 
      printf("\n\n ***************** Sparse Vector: %s *******************\n", title);
    else 
      printf("\n\n ***************** Vector: %s *******************\n", title);
    fflush(stdout);
  }

  MPI_Barrier(WORLD);

  for (pn=0; pn<tot_nproc; ++pn) {

    if ( my_pnum == pn ) {
      /***
      printf("\n [%2d] ", pn); 
      ***/
      
      for (i=0; i<tot_nproc; ++i) 
	if (kind == 'O' || kind == 'o')
	  printf("%4d ", array[i]/2); /* OSC scheme */
	else
	  printf("%4d ", array[i]); /* TSC scheme */
      /***
      printf(" : %d", total);
      ***/
      printf("\n");
      fflush(stdout);
    }
    MPI_Barrier(WORLD);
  }

  MPI_Barrier(WORLD);
  if (only_master) 
    printf("\n");


}



/******************************************************************************
 *                                                                            *
 *  Function : print_segment()
 *                                                                            *
 ******************************************************************************/


void print_segment(B, size, list)
     int B, size, *list;
{
  int pn, i;

  MPI_Barrier(WORLD);
  if (only_master) {
    printf("\n\n ******************** Segment Scan ********************\n");
  }

  MPI_Barrier(WORLD);

  for (pn=0; pn<tot_nproc; ++pn) {

    if ( my_pnum == pn ) {
      printf("\n  ==============> [PN:%3d] B = %d\n", pn, B); 
      if (B != -1) {
	printf("   [Size = %3d]", size);
	for (i=0; i<size; ++i)
	  printf("%3d",list[i]);
      printf("\n");
      fflush(stdout);
      }
    }

    MPI_Barrier(WORLD);

  }

  if (my_pnum == 0 )
    printf("\n\n");
 
  MPI_Barrier(WORLD);

}
