/********************************************************************************

 Programmed by Khaled Alsabti

 Modified by SEUNG-JO BAE

  - Goal: 

  - Algorithm: 

  - Input: 

  - Output: 

  - Version: 1.2

  - Communication: 

  - Note: 1) Distributed version = sample_data_1d

             a) One-dimensional array (source and target array)  
             b) Equal sized sets of skew dtat for hot processors 
	     c) Source & Target distribution: Block              

	  2) CS Distributed version = cs_sample_data_1d (7/18/96)

	  3) hot_num =     percentage
                               -or-
                        number of hot processor      (7/21/96)

	  4) Combine sample_data_1d with cs_sample_data_1d (7/21/96)

	  5) If hot_degree < 0 or > 1, randomly generate hot_degree and
             hot_num on each processor (7/21/96)

          6) For exact timing, make processor #0 to be the first hot processor
             That is, hot_proc[0] = 0. (7/31/96)

	  7) Define hot_range such that the local index of hot pointers on
             a hot processor: 0 <=   < hot_range * t_L (1/6/97)


	  2-1) Use t_L_table instead of t_L (7/12/97)
	  
********************************************************************************/

#include "global.h"

/******************************************************************************
 *                                                                            *
 *  Function : sample_data_1d                                                 *
 *                                                                            *
 ******************************************************************************/

void sample_data_1d (data, pointer, target, hot_degree, hot_fnum, hot_proc, 
		     hot_range, seed, pr_control)
     Data_Type data[Lmax];                /* Source data array */
     int pointer[Lmax];             /* Source pointer array */
     Data_Type target[Lmax];              /* Target array */
     double hot_degree;             /* Degree of skew */
     double hot_fnum;               /* Number of hot processors or percentage */
     int   hot_proc[Hot_Proc_Max];  /* hot_num Hot processors */
     double hot_range;		    /* range of hot pointers in local array */
     int seed;
     int pr_control;
{ 
  int hot_num;                      /* Number of hot processors */
  int i, j, ridx, bound, pn, lptr;
  int hot_size[Hot_Proc_Max],       /* Size of skew pointers */
      random_size;                  /* Size of random pointers */
  int hot_start[Hot_Proc_Max],      /* Starting pointer of skew data */
      hot_tot_size;
  int d, m;			    /* d=div, m=mod */
  int hpn;

  Boolean random_hot_degree;

  Boolean same;
  int s;
  int ori_seed;

  ori_seed = seed;

  /*----------------------- Calculate hot_num ------------------------------*/

  if ( fmod(hot_fnum, 1.) == 0. )
    hot_num = (int) hot_fnum;
  else
    hot_num = (int) (t_tot_nproc * hot_fnum);

  /*------------------------- Check arguments -----------------------------*/

  if ( hot_num <= 0 || hot_num > t_tot_nproc/2 )       {
    printf("\n Improper hot_num %d. Exit! \n", hot_num);
    exit(0);
  };

  /*----------------- randomly selected hot_degree and hot_num -------------*/

  random_hot_degree = ( (hot_degree < 0.) || (hot_degree > 1.) );
  if ( random_hot_degree ) { 
    do {
      srand48(my_pnum * 23 + seed);
      for (i=0; i<my_pnum*5; ++i)
	lrand48();
      hot_degree = (double) ((double) lrand48() / (double) Max_long);
      if ( (hot_degree < 0.) || (hot_degree > 1.) ) {
	printf("\n Improper random hot_degree = %f (PN=%d). Exit! \n", 
	       hot_degree, my_pnum);
	exit(0);
      }
      hot_num = lrand48() % (t_tot_nproc/2);
      seed++;
    } while (hot_num == 0 && hot_degree != 0.);
  }

  /*------------------ Initialize seed for random number -------------------*/

  srand48(my_pnum * 100 + seed);
  seed++;    

/***
CMMD_fset_io_mode(stdout,CMMD_sync_bc);
printf("\n............... 1 .............");
fflush(stdout);
CMMD_sync_with_nodes();
***/


  /*--------------------- Generate random pointers -------------------------*/

  random_size = s_L[0] * (1.- hot_degree);

/***
printf("\n............... 2.0 .............");
fflush(stdout);
CMMD_sync_with_nodes();
***/

  for (ridx=0; ridx < random_size ; ridx++)
    pointer[ridx] = lrand48() % t_N[0];

/***
printf("\n............... 2 .............");
fflush(stdout);
CMMD_sync_with_nodes();
***/
    

  /*------------------------ Generate hot pointers -------------------------*/

  hot_tot_size = 0;
  d = ( s_L[0] - random_size ) / hot_num; 
  m = ( s_L[0] - random_size ) % hot_num; 
  for (i=0; i < hot_num ; i++)  {    /* almost equal sized sets */
    hot_size[i] = d;
    hot_tot_size += hot_size[i];
  }

  if (m) 
    for (i=0; i < m ; i++)  {    /* almost equal sized sets */
      ++ hot_size[i];
      ++ hot_tot_size;
  }
    
  if ( random_size + hot_tot_size < s_L[0] ) {
    printf("\n random_size %d + hot_tot_size %d < s_L[0] %d . Exit! \n", 
	   random_size, hot_tot_size, s_L[0]);
    exit(0);
  }

  if ( random_hot_degree ) 
    srand48(my_pnum * 27 + seed);
  else
    srand48(seed++);               /* every proc has the same hot processors */

/*-------------------- 7/31/96 ------------------*/
/*--  for (i=0; i < hot_num ; i++) {           --*/
/*--                                           --*/
/*--    hot_proc[i] = lrand48() % t_tot_nproc; --*/
/*-------------------- 7/31/96 ------------------*/

  /*--- GENERAL ---*/
  hot_proc[0] = -1;
  hpn = t_tot_nproc-1;
  while ( hot_proc[0] < 0 ) {
    if (t_L_table[0][hpn])
      hot_proc[0] = hpn;
    else
      -- hpn;
  }
  /* hot_proc[0] = 0; */

/***
CMMD_fset_io_mode(stdout,CMMD_sync_bc);
printf("\n ****** hpn0 = %d, size = %d\n", hpn,t_L_table[0][hpn]);
fflush(stdout);
CMMD_sync_with_nodes();
***/

  for (i=1; i < hot_num ; i++) {

    hot_proc[i] = lrand48() % t_tot_nproc;

    /* Each hot processor should be different */
    if ( i > 0 ) {

      same = false;
      for (s=0; s < i ; ++s)
	same = same || ( hot_proc[s] == hot_proc[i] );
      
      while ( same ) {
	hot_proc[i] = lrand48() % t_tot_nproc;
	same = false;
	for (s=0; s < i ; ++s)
	  same = same || ( hot_proc[s] == hot_proc[i] );
      }

    } /* if */

  } /* for */

  srand48(my_pnum * 100 + seed); 

  /*--- GENERAL ---*/
  bound = random_size;
  for (i=0; i < hot_num ; i++) {
    bound += hot_size[i];
    for(; ridx < bound; ridx++) {
  	lptr = lrand48() % (int)(t_L_table[0][hot_proc[i]] * hot_range);
  	pointer[ridx] = map_to_global(TARGET, hot_proc[i], lptr, 0);
    }
  }

/***
printf("\n............... 3 .............");
fflush(stdout);
CMMD_sync_with_nodes();
***/


  /*----------------------- Initialize source  array ---------------------*/
  for(j=0; j < s_L[0] ; j++) 
    data[j] = 1;
      
/***
printf("\n............... 4 .............");
fflush(stdout);
CMMD_sync_with_nodes();
***/


  /*----------------------- Print out data information --------------------*/
 
  if (pr_control==0) {

    if ( random_hot_degree ) { 

      MPI_Barrier(WORLD);

      for (pn=0; pn<tot_nproc; ++pn) {
      
	if ( my_pnum == pn ) {

	  printf("\n********* Sample Data [PN:%3d] **************\n",
		 my_pnum);
	  printf(" Degree of skew          = %f\n", hot_degree);
	  printf(" Size of random pointers = %d\n\n", random_size);
	  printf(" Num of hot processors   = %d\n", hot_num);
	  for (i=0; i < hot_num ; i++) {
	    printf("\n     Hot proc #%2d          : %d\n", i, hot_proc[i]);
	    printf("     Size of skew pointers : %d\n",hot_size[i]);
	  }
	  fflush(stdout);

	}
	MPI_Barrier(WORLD);
      }

      if (only_master) {
	printf("*********************************************\n");
	fflush(stdout);
      }
    }

    else if (only_master) {
      printf("\n************* Sample Data *******************\n");
      printf(" Random seed             = %d\n", ori_seed);
      printf(" Degree of skew          = %f\n", hot_degree);
      printf(" Size of random pointers = %d\n\n", random_size);
      printf(" Range of hot pointers   = %f\n", hot_range);
      printf(" Num of hot processors   = %d\n", hot_num);
      for (i=0; i < hot_num ; i++) {
	printf("\n     Hot proc #%2d          : %d\n", i, hot_proc[i]);
	printf("     Size of skew pointers : %d\n",hot_size[i]);
      }
      printf("*********************************************\n");
    }
  }

}




